#!/usr/bin/env python
import logging
from datetime import datetime
from datetime import timezone
from typing import ClassVar
from typing import Optional

from .base import BaseElement
from .base import NamedBaseElement
from .base import ValuedBaseElement
from caldav.lib.namespace import ns

utc_tz = timezone.utc


def _to_utc_date_string(ts):
    # type (Union[date,datetime]]) -> str
    """coerce datetimes to UTC (assume localtime if nothing is given)"""
    if isinstance(ts, datetime):
        try:
            ## for any python version, this should work for a non-native
            ## timestamp.
            ## in python 3.6 and higher, ts.astimezone() will assume a
            ## naive timestamp is localtime (and so do we)
            ts = ts.astimezone(utc_tz)
        except:
            ## native time stamp and the current python version is
            ## not able to treat it as localtime.
            import tzlocal

            ts = ts.replace(tzinfo=tzlocal.get_localzone())

            mindate = datetime.min.replace(tzinfo=utc_tz)
            maxdate = datetime.max.replace(tzinfo=utc_tz)
            if mindate + ts.tzinfo.utcoffset(ts) > ts:
                logging.error(
                    "Cannot coerce datetime %s to UTC. Changed to min-date.", ts
                )
                ts = mindate
            elif ts > maxdate - ts.tzinfo.utcoffset(ts):
                logging.error(
                    "Cannot coerce datetime %s to UTC. Changed to max-date.", ts
                )
                ts = maxdate
            else:
                ts = ts.astimezone(utc_tz)

    return ts.strftime("%Y%m%dT%H%M%SZ")


# Operations
class CalendarQuery(BaseElement):
    tag: ClassVar[str] = ns("C", "calendar-query")


class FreeBusyQuery(BaseElement):
    tag: ClassVar[str] = ns("C", "free-busy-query")


class Mkcalendar(BaseElement):
    tag: ClassVar[str] = ns("C", "mkcalendar")


class CalendarMultiGet(BaseElement):
    tag: ClassVar[str] = ns("C", "calendar-multiget")


class ScheduleInboxURL(BaseElement):
    tag: ClassVar[str] = ns("C", "schedule-inbox-URL")


class ScheduleOutboxURL(BaseElement):
    tag: ClassVar[str] = ns("C", "schedule-outbox-URL")


# Filters
class Filter(BaseElement):
    tag: ClassVar[str] = ns("C", "filter")


class CompFilter(NamedBaseElement):
    tag: ClassVar[str] = ns("C", "comp-filter")


class PropFilter(NamedBaseElement):
    tag: ClassVar[str] = ns("C", "prop-filter")


class ParamFilter(NamedBaseElement):
    tag: ClassVar[str] = ns("C", "param-filter")


# Conditions
class TextMatch(ValuedBaseElement):
    tag: ClassVar[str] = ns("C", "text-match")

    def __init__(self, value, collation: str = "i;octet", negate: bool = False) -> None:
        super(TextMatch, self).__init__(value=value)

        if self.attributes is None:
            raise ValueError("Unexpected value None for self.attributes")

        self.attributes["collation"] = collation
        if negate:
            self.attributes["negate-condition"] = "yes"


class TimeRange(BaseElement):
    tag: ClassVar[str] = ns("C", "time-range")

    def __init__(
        self, start: Optional[datetime] = None, end: Optional[datetime] = None
    ) -> None:
        ## start and end should be an icalendar "date with UTC time",
        ## ref https://tools.ietf.org/html/rfc4791#section-9.9
        super(TimeRange, self).__init__()

        if self.attributes is None:
            raise ValueError("Unexpected value None for self.attributes")

        if start is not None:
            self.attributes["start"] = _to_utc_date_string(start)
        if end is not None:
            self.attributes["end"] = _to_utc_date_string(end)


class NotDefined(BaseElement):
    tag: ClassVar[str] = ns("C", "is-not-defined")


# Components / Data
class CalendarData(BaseElement):
    tag: ClassVar[str] = ns("C", "calendar-data")


class Expand(BaseElement):
    tag: ClassVar[str] = ns("C", "expand")

    def __init__(
        self, start: Optional[datetime], end: Optional[datetime] = None
    ) -> None:
        super(Expand, self).__init__()

        if self.attributes is None:
            raise ValueError("Unexpected value None for self.attributes")

        if start is not None:
            self.attributes["start"] = _to_utc_date_string(start)
        if end is not None:
            self.attributes["end"] = _to_utc_date_string(end)


class Comp(NamedBaseElement):
    tag: ClassVar[str] = ns("C", "comp")


# Uhhm ... can't find any references to calendar-collection in rfc4791.txt
# and newer versions of baikal gives 403 forbidden when this one is
# encountered
# class CalendarCollection(BaseElement):
#     tag = ns("C", "calendar-collection")


# Properties
class CalendarUserAddressSet(BaseElement):
    tag: ClassVar[str] = ns("C", "calendar-user-address-set")


class CalendarUserType(BaseElement):
    tag: ClassVar[str] = ns("C", "calendar-user-type")


class CalendarHomeSet(BaseElement):
    tag: ClassVar[str] = ns("C", "calendar-home-set")


# calendar resource type, see rfc4791, sec. 4.2
class Calendar(BaseElement):
    tag: ClassVar[str] = ns("C", "calendar")


class CalendarDescription(ValuedBaseElement):
    tag: ClassVar[str] = ns("C", "calendar-description")


class CalendarTimeZone(ValuedBaseElement):
    tag: ClassVar[str] = ns("C", "calendar-timezone")


class SupportedCalendarComponentSet(ValuedBaseElement):
    tag: ClassVar[str] = ns("C", "supported-calendar-component-set")


class SupportedCalendarData(ValuedBaseElement):
    tag: ClassVar[str] = ns("C", "supported-calendar-data")


class MaxResourceSize(ValuedBaseElement):
    tag: ClassVar[str] = ns("C", "max-resource-size")


class MinDateTime(ValuedBaseElement):
    tag: ClassVar[str] = ns("C", "min-date-time")


class MaxDateTime(ValuedBaseElement):
    tag: ClassVar[str] = ns("C", "max-date-time")


class MaxInstances(ValuedBaseElement):
    tag: ClassVar[str] = ns("C", "max-instances")


class MaxAttendeesPerInstance(ValuedBaseElement):
    tag: ClassVar[str] = ns("C", "max-attendees-per-instance")


class Allprop(BaseElement):
    tag: ClassVar[str] = ns("C", "allprop")


class ScheduleTag(BaseElement):
    tag: ClassVar[str] = ns("C", "schedule-tag")



import json
import logging
import os

"""
This configuration parsing code was just copied from my plann library (and will be removed from there at some point in the future).  Test coverage is poor as for now.
"""

## This is being moved from my plann library.  The code itself will be introduced into caldav 2.0, but proper test code and documentation will come in a later release (2.1?)


## TODO TODO TODO - write test code for all the corner cases
## TODO TODO TODO - write documentation of config format
def expand_config_section(config, section="default", blacklist=None):
    """
    In the "normal" case, will return [ section ]

    We allow:

    * * includes all sections in config file
    * "Meta"-sections in the config file with the keyword "contains" followed by a list of section names
    * Recursive "meta"-sections
    * Glob patterns (work_* for all sections starting with work_)
    * Glob patterns in "meta"-sections
    """
    ## Optimizating for a special case.  The results should be the same without this optimization.
    if section == "*":
        return [x for x in config if not config[x].get("disable", False)]

    ## If it's not a glob-pattern ...
    if set(section).isdisjoint(set("[*?")):
        ## If it's referring to a "meta section" with the "contains" keyword
        if "contains" in config[section]:
            results = []
            if not blacklist:
                blacklist = set()
            blacklist.add(section)
            for subsection in config[section]["contains"]:
                if not subsection in results and not subsection in blacklist:
                    for recursivesubsection in expand_config_section(
                        config, subsection, blacklist
                    ):
                        if not recursivesubsection in results:
                            results.append(recursivesubsection)
            return results
        else:
            ## Disabled sections should be ignored
            if config.get("section", {}).get("disable", False):
                return []

            ## NORMAL CASE - return [ section ]
            return [section]
    ## section name is a glob pattern
    matching_sections = [x for x in config if fnmatch(x, section)]
    results = set()
    for s in matching_sections:
        if set(s).isdisjoint(set("[*?")):
            results.update(expand_config_section(config, s))
        else:
            ## Section names shouldn't contain []?* ... but in case they do ... don't recurse
            results.add(s)
    return results


def config_section(config, section="default"):
    if section in config and "inherits" in config[section]:
        ret = config_section(config, config[section]["inherits"])
    else:
        ret = {}
    if section in config:
        ret.update(config[section])
    return ret


def read_config(fn, interactive_error=False):
    if not fn:
        cfgdir = f"{os.environ.get('HOME', '/')}/.config/"
        for config_file in (
            f"{cfgdir}/caldav/calendar.conf",
            f"{cfgdir}/caldav/calendar.yaml",
            f"{cfgdir}/caldav/calendar.json",
            f"{cfgdir}/calendar.conf",
            "/etc/calendar.conf",
            "/etc/caldav/calendar.conf",
        ):
            cfg = read_config(config_file)
            if cfg:
                return cfg
        return None

    ## This can probably be refactored into fewer lines ...
    try:
        try:
            with open(fn, "rb") as config_file:
                return json.load(config_file)
        except json.decoder.JSONDecodeError:
            ## Late import, wrapped in try/except.  yaml is external module,
            ## and not included in the requirements as for now.
            try:
                import yaml

                try:
                    with open(fn, "rb") as config_file:
                        return yaml.load(config_file, yaml.Loader)
                except yaml.scanner.ScannerError:
                    logging.error(
                        f"config file {fn} exists but is neither valid json nor yaml.  Check the syntax."
                    )
            except ImportError:
                logging.error(
                    f"config file {fn} exists but is not valid json, and pyyaml is not installed."
                )

    except FileNotFoundError:
        ## File not found
        logging.info("no config file found")
    except ValueError:
        if interactive_error:
            logging.error(
                "error in config file.  Be aware that the interactive configuration will ignore and overwrite the current broken config file",
                exc_info=True,
            )
        else:
            logging.error("error in config file.  It will be ignored", exc_info=True)
    return {}



#!/usr/bin/env python
import logging

try:
    from ._version import __version__
except ModuleNotFoundError:
    __version__ = "(unknown)"
    import warnings

    warnings.warn(
        "You need to install the `build` package and do a `python -m build` to get caldav.__version__ set correctly"
    )
from .davclient import DAVClient
from .davclient import get_davclient
from .search import CalDAVSearcher

## TODO: this should go away in some future version of the library.
from .objects import *

## We should consider if the NullHandler-logic below is needed or not, and
## if there are better alternatives?
# Silence notification of no default logging handler
log = logging.getLogger("caldav")


class NullHandler(logging.Handler):
    def emit(self, record) -> None:
        pass


log.addHandler(NullHandler())

__all__ = ["__version__", "DAVClient", "get_davclient"]
